# frozen_string_literal: true

require "qi"

require_relative "error/game_over/illegal_move/piece_not_found"
require_relative "error/game_over/illegal_move/square_not_found"

module Sanqi
  # The position abstraction.
  class Position
    def self.starting_position_meth
      {
        CHESS:       {
          CHESS:       :chess_vs_chess,
          PUCHI_SHOGI: :chess_vs_puchi_shogi,
          XIAOQI:      :chess_vs_xiaoqi
        },
        PUCHI_SHOGI: {
          CHESS:       :puchi_shogi_vs_chess,
          PUCHI_SHOGI: :puchi_shogi_vs_puchi_shogi,
          XIAOQI:      :puchi_shogi_vs_xiaoqi
        },
        XIAOQI:      {
          CHESS:       :xiaoqi_vs_chess,
          PUCHI_SHOGI: :xiaoqi_vs_puchi_shogi,
          XIAOQI:      :xiaoqi_vs_xiaoqi
        }
      }
    end
    private_class_method :starting_position_meth

    def self.between(south_variant, north_variant)
      public_send(starting_position_meth.fetch(south_variant).fetch(north_variant))
    end

    def self.chess_vs_chess
      new(turn_to_north: false, square: CHESS_VS_CHESS_SQUARE)
    end

    def self.chess_vs_puchi_shogi
      new(turn_to_north: false, square: CHESS_VS_PUCHI_SHOGI_SQUARE)
    end

    def self.chess_vs_xiaoqi
      new(turn_to_north: false, square: CHESS_VS_XIAOQI_SQUARE)
    end

    def self.puchi_shogi_vs_chess
      new(turn_to_north: false, square: PUCHI_SHOGI_VS_CHESS_SQUARE)
    end

    def self.puchi_shogi_vs_puchi_shogi
      new(turn_to_north: false, square: PUCHI_SHOGI_VS_PUCHI_SHOGI_SQUARE)
    end

    def self.puchi_shogi_vs_xiaoqi
      new(turn_to_north: false, square: PUCHI_SHOGI_VS_XIAOQI_SQUARE)
    end

    def self.xiaoqi_vs_chess
      new(turn_to_north: false, square: XIAOQI_VS_CHESS_SQUARE)
    end

    def self.xiaoqi_vs_puchi_shogi
      new(turn_to_north: false, square: XIAOQI_VS_PUCHI_SHOGI_SQUARE)
    end

    def self.xiaoqi_vs_xiaoqi
      new(turn_to_north: false, square: XIAOQI_VS_XIAOQI_SQUARE)
    end

    def self.squares_count
      SQUARE_COUNT
    end

    attr_reader :square

    # Initialize a position.
    #
    # @param turn_to_north [Boolean] The identifier of the player who must play.
    # @param square [Hash] The index of each piece on the board.
    #
    # @example Dump a classic Tsume Shogi problem
    #   new(
    #     "turn_to_north": false,
    #     "square": {
    #        3 => "s",
    #        4 => "k",
    #        5 => "s",
    #       22 => "+P",
    #       43 => "+B"
    #     }
    #   )
    def initialize(turn_to_north:, square:)
      @turn_to_north = turn_to_north
      @square = square
    end

    def after(source, target, new_piece, captured_piece = nil)
      hash = ::Qi.call(source, target, new_piece, captured_piece, square: @square, in_hand: [])
      hash.delete(:in_hand)

      self.class.new(turn_to_north: turn_to_south?, **hash)
    end

    def piece(index)
      pieces.fetch(index) { raise Error::GameOver::IllegalMove::PieceNotFound }
    end

    def pieces
      @square
    end

    def promotion_square?(index)
      promotion_square_ids.include?(index)
    end

    # @return [Boolean] Is turn to north?
    def turn_to_north?
      @turn_to_north
    end

    # @return [Boolean] Is turn to south?
    def turn_to_south?
      !turn_to_north?
    end

    # @param conditions [Hash] A hash of integers and symbols.
    #
    # @example A hash of square patterns
    #   { 10 => nil, 20 => nil, 30 => nil, 40 => nil, 50 => nil }
    #
    # @return [Boolean] Is the position matching the given pattern?
    def match?(conditions)
      conditions.all? { |index, state| correct?(index, state) }
    end

    # @param index [Integer] The square ID.
    #
    # @return [String, nil] The square content.
    def square_content(index)
      @square.fetch(index, nil)
    end

    # @note Faster than FEEN to compute.
    def inspect
      inspect_square = @square.keys.sort.inject("") { |s, key| s + "#{key}:#{@square.fetch(key)}" }
      inspect_turn = (turn_to_north? ? "north-turn" : "south-turn")

      "<Sanqi::Position #{inspect_square} #{inspect_turn}>"
    end

    protected

    # @param square_id [Integer] The square ID.
    # @param state [Symbol] The state of the square.
    #
    # @return [Boolean] Is the state of the square correct?
    def correct?(square_id, state)
      content = square_content(square_id)

      case state
      when :occupied
        !content.nil?
      when :enemy
        if turn_to_north?
          !content.nil? && content.upcase == content
        else
          !content.nil? && content.downcase == content
        end
      else
        state == content
      end
    end

    def promotion_square_ids
      if turn_to_north?
        north_promotion_square_ids
      else
        south_promotion_square_ids
      end
    end

    def north_promotion_square_ids
      (54..80)
    end

    def south_promotion_square_ids
      (0..26)
    end
  end
end
